"""
Tools to encapsulate generic receive connectors
"""
import logging
import paho.mqtt.client as mqtt

from .decorators import ensure_connection
from .topic import TopicCollection


class Receiver(mqtt.Client):
    """
    Receive messages from specific mqtt queue

    Args:
        topics (list):
            list of strings with the topic names

        url (str):
            mqtt server url

        port (int):
            mqtt server port

        keepalive (int):
            mqtt keepalive time in seconds
    """
    def __init__(self, topics=[], url=None, port=None, keepalive=None):
        super().__init__()
        self.input_topics = TopicCollection(topics)
        self.url = url
        self.port = port
        self.keepalive = keepalive

    @staticmethod  # decorator to avoid double 'self' on paho callback
    def on_connect(self, userdata, flags, rc):
        """
        The callback for when the client receives a CONNACK response
        from the server.
        """
        logging.info(f'Connected with Mosquitto Server: (code) {rc}')

    @staticmethod  # decorator to avoid double 'self' on paho callback
    def on_message(self, userdata, message):
        """
        The callback for when a PUBLISH message is received from the server.
        """
        logging.debug(f"Message arrived from {message.topic}")
        self.process(userdata, message)

    def process(self, userdata, message):
        raise NotImplementedError

    def run(self):
        """
        Blocking call that processes network traffic, dispatches callbacks and
        handles reconnecting.
        Other loop*() functions are available that give a threaded interface
        and a manual interface.
        """
        self.listen(self.input_topics)

        logging.info('Getting into the listening loop')
        self.running = True
        while self.running:
            self.loop()

    @ensure_connection
    def listen(self, topics):
        """
        Subscribe to a list of channels

        Args:
            topics (list):
                list of topics to subscribe the mqtt listener
        """
        logging.debug(f'Listen to {list(map(lambda x: x.name, topics))}')

        for topic in map(lambda x: x.name, topics):
            try:
                self.subscribe(topic)
                logging.debug(f'Subscribed the {topic} topic')
            except Exception:
                logging.debug(f"Can't subscribe the {topic} topic")

    def stop(self):
        """
        Stop running
        """
        self.running = False



import logging
import paho.mqtt.client as mqtt

from .message import Message
from .decorators import ensure_connection
from .topic import TopicCollection


class Transmitter(mqtt.Client):
    """
    Handler for MQTT publishing

    Args:
        topics (dict):
            keys identify the topic,s for the mqtt topic
            names to publish
    """
    def __init__(self, topics=[], url=None, port=None, keepalive=None):
        self.output_topics = TopicCollection(topics)
        self.url = url
        self.port = port
        self.keepalive = keepalive

    @staticmethod
    def on_connect(self, userdata, flags, rc):
        """
        The callback for when the client receives a CONNACK response
        from the server.
        """
        logging.info(f'Connected with MQTT Server: (code) {rc}')

    @staticmethod
    def on_disconnect(self, userdata, rc):
        """
        Called when the client disconnects from the broker.
        """
        self.connected = False

    @ensure_connection
    def send(self, message):
        """
        Publish string to the 2RSystem queue

        Args:
            data (str):
                string message to publish
            to (str):
                topic name
        """
        receivers = []
        if isinstance(message, Message):
            receivers = message.filter_topics(self.output_topics)

        logging.debug(f"Sending message to {receivers}")

        for topic in map(lambda x: x.name, receivers):
            logging.info(f'Publishing on topic {topic}')
            self.publish(topic, message.encoded)



import logging
from multiprocessing import Process

from .gabby import Gabby


class Controller:
    _instance = None

    def __new__(cls, thread=False):
        """
        Singleton implementation
        """
        if Controller._instance is None:
            obj = object.__new__(cls)
            Controller._instance = obj
            Controller._instance.gabbies = []
        return Controller._instance

    def add_gabby(self, gabby):
        if isinstance(gabby, Gabby):
            self.gabbies.append(gabby)
        else:
            raise TypeError('You can only add Gabby objects')

    def run(self):
        """
        Run a process for each 2RSystem sub module
        """
        runner_procs = [self._run_instance(x) for x in self.gabbies]

        for proc in runner_procs:
            proc.join()

        logging.info("[Success] Shutted down")

    def _run_instance(self, gabby):
        """
        Start a new process to execute a given function
        """
        logging.info(f'Starting module')
        proc = Process(target=gabby.run)
        proc.start()
        return proc



"""
Gabby module witch creates the Gabby class to handle creation of
message queue nodes for intercommunication
"""
import logging

from .transmitter import Transmitter
from .receiver import Receiver
from .message import Message


class Gabby(Transmitter, Receiver):
    def __init__(self, input_topics=None, output_topics=None, decode_input=True,
                 url=None, port=None, keepalive=None):
        """
        Processor initializer

        Args:
            topics (collection):
                output topics to send results

            decoder (bool), default=True:
                enable auto encoding/decoding of any received message
        """
        Receiver.__init__(self, input_topics or [], url, port, keepalive)
        Transmitter.__init__(self, output_topics or [], url, port, keepalive)
        self.decode_input = decode_input

    def process(self, userdata, message):
        if self.decode_input:
            topic_name = message.topic
            topics = list(
                filter(
                    lambda x: x.name == topic_name,
                    self.input_topics
                )
            )
            message = Message.decode(message.payload, topics)

        logging.debug(f'Processing message: {message}')
        response_messages = self.transform(message)
        for msg in response_messages:
            self.send(msg)

    def transform(self, message):
        """
        Abstract method to process any received message

        Args:
            message (Message or paho.mqtt.MQTTMessage):
                message from queue decoded or not depending on 'decode' var

            author (str):
                message's writer

        Return:
            Collection of messages to be transmitted or an empty list
        """
        return [Message(message.data, self.output_topics)]




"""
Module to handle message serialization
"""
import logging
import struct

from .topic import TopicCollection


class Message:
    """
    Handler for creating complex data structs to transmit and process
    through the MQTT pipelines
    """
    def __init__(self, data, topics=[], fmt=None):
        logging.debug(f'Creating a new message with this data: {data}')
        self.data = data
        self.topics = TopicCollection(topics)
        self.fmt = topics[0].fmt if topics else fmt

        if self.fmt is None:
            raise ValueError("You should setup the topic or the fmt")

    @property
    def encoded(self):
        """
        Converts the data attribute to a complex bytestring to send
        through the MQTT broker.
        """
        return struct.pack(self.fmt, *self.data)

    def __str__(self):
        return str(self.data)

    def filter_topics(self, topics):
        """
        It filters topics that match the same format of the message.

        Args:
            topics (collection):
                list of topics to filter matches
        """
        if not self.topics:
            return filter(lambda x: x.fmt == self.fmt, topics)
        else:
            return filter(lambda x: x in self.topics, topics)

    def belongs_to(self, name):
        """
        Check if an name matches to any self.topics name
        """
        return self.topics.find_by(name=name) is not None

    @staticmethod
    def decode(message, topics=[], fmt=None):
        """
        Convert an MQTTMessage to a Message

        Args:
            message (MQTTMessage):
                A paho.mqtt.MQTTMessage received from any message queue

            topics (collection):
                list of topics to extract fmt of the first topic

            fmt (str):
                struct model of encoded message
        """
        if fmt or topics:
            data = struct.unpack(topics[0].fmt if topics else fmt, message)
            logging.debug(f'Decoded data: {data}')
            return Message(data, topics=topics, fmt=fmt)
        else:
            raise ValueError("You should setup the topic or the fmt")



"""
Decorators module
"""
from .settings import URL, PORT, KEEPALIVE


def ensure_connection(func):
    """
    Decorator to ensure connection before execute any decorated method
    """
    def wrapper(self, *args, **kwargs):
        try:
            getattr(self, "connected")
        except AttributeError:
            self.connected = False
        finally:
            if not self.connected:
                self.connect(
                    self.url or URL,
                    self.port or PORT,
                    self.keepalive or KEEPALIVE
                )
                self.connected = True
        return func(self, *args, **kwargs)
    return wrapper



from collections import namedtuple


Topic = namedtuple('Topic', ['name', 'fmt'])


class TopicCollection(list):
    """
    Collection to handler Topic groups operations
    """
    def __init__(self, *args):
        if args and isinstance(args[0], Topic):
            self.extend(args)
        else:
            self.extend(args[0])

    def append(self, topic):
        if isinstance(topic, Topic):
            if list(filter(lambda x: x.name == topic.name, self)) == []:
                super().append(topic)
            else:
                raise AttributeError("Topic must be unique")
        else:
            raise TypeError("You can only add Topic objects")

    def extend(self, topics):
        for t in topics:
            self.append(t)

    def find_by(self, **kwargs):
        """
        Get topic by any attribute.

        Return:
            First that matches

        Example:
            >>> TopicCollection(Topic('a', 'b')).find_by(name='a')
            ... Topic(name='a', fmt='b')
        """
        match_topics = self.filter_by(**kwargs)
        return match_topics[0] if match_topics else None

    def filter_by(self, **kwargs):
        """
        Filter by any topic attribute

        Example:
            >>> TopicCollection(Topic('a', 'b')).filter_by(name='a')
            ... [Topic(name='a', fmt='b')]
        """
        return list(filter(
            lambda x: all([getattr(x, attr) == value
                           for attr, value in kwargs.items()]), self
        ))
